/* Defines ******************************************************************/

#define _CRT_SECURE_NO_WARNINGS

/* Includes *****************************************************************/

#include <windows.h>
#include <tchar.h>
#include <stdio.h>
#include <direct.h>

/* Constants ****************************************************************/

LPCTSTR title = _T("SimCity 3000 Resolution Patcher 1.1 by ToastyX");

const struct edition
{
	LPCTSTR name;
	LPCTSTR exe_dir;
	LPCTSTR exe_file;
}
editions[] =
{
	{_T("SimCity 3000 Unlimited"),     _T("Apps"), _T("sc3U.exe")},
	{_T("SimCity 3000 World Edition"), _T("Apps"), _T("sc3U.exe")},
	{_T("SimCity 3000 UK Edition"),    _T("Apps"), _T("sc3U.exe")},
	{_T("SimCity 3000 Deutschland"),   _T("Apps"), _T("sc3U.exe")},
	{_T("SimCity 3000"),               _T("Game"), _T("SC3.EXE")},
	{NULL}
};

const unsigned max_file_size = 3000000;

/* Functions ****************************************************************/

int error_message(LPCTSTR text)
{
	MessageBox(NULL, text, _T("Error"), MB_ICONERROR);
	return 1;
}

/****************************************************************************/

BOOL strip_quotes(LPTSTR text)
{
	LPTSTR src;
	LPTSTR dst;

	for (src = dst = text; *src; src++)
	{
		if (*src != _T('"'))
		{
			*dst = *src;
			dst++;
		}
	}

	*dst = 0;
	return TRUE;
}

/****************************************************************************/

BOOL strip_file_name(LPTSTR path_name)
{
	LPTSTR file_name;

	file_name = _tcsrchr(path_name, _T('\\'));

	if (!file_name)
		return FALSE;

	file_name++;
	*file_name = 0;
	return TRUE;
}

/****************************************************************************/

unsigned patch(unsigned char *buffer, unsigned size)
{
	unsigned index;
	unsigned offset;
	unsigned *width;
	unsigned *height;
	unsigned found = 0;

	if (size < 16)
		return 0;

	for (index = 0; index < size - 16; index++)
	{
		width = (unsigned *)&buffer[index];

		if (*width == 1152)
		{
			for (offset = 4; offset < 12; offset++)
			{
				height = (unsigned *)&buffer[index + offset];

				if (*height == 864 && buffer[index + offset + 4] != 0xEB)
				{
					*width = GetSystemMetrics(SM_CXSCREEN);
					*height = GetSystemMetrics(SM_CYSCREEN);
					found++;
				}
			}
		}
	}

	return found;
}

/****************************************************************************/

BOOL readable_file(LPCTSTR src_file)
{
	FILE *file;

	file = _tfopen(src_file, _T("rb"));

	if (!file)
		return FALSE;

	fclose(file);
	return TRUE;
}

/****************************************************************************/

BOOL replace_file(LPCTSTR src_file, LPCTSTR dst_file, LPCTSTR bak_file)
{
	LPCTSTR tmp_file;

	tmp_file = _ttmpnam(NULL);

	if (!tmp_file)
		return FALSE;

	if (!bak_file || !readable_file(dst_file))
	{
		_trename(dst_file, tmp_file);

		if (_trename(src_file, dst_file) != 0)
		{
			_trename(tmp_file, dst_file);
			return FALSE;
		}
	}
	else
	{
		_trename(bak_file, tmp_file);

		if (_trename(dst_file, bak_file) != 0)
		{
			_trename(tmp_file, bak_file);
			return FALSE;
		}

		if (_trename(src_file, dst_file) != 0)
		{
			_trename(bak_file, dst_file);
			_trename(tmp_file, bak_file);
			return FALSE;
		}
	}

	_tunlink(tmp_file);
	return TRUE;
}

/****************************************************************************/

BOOL write_file(unsigned char *src_buffer, unsigned src_size, LPCTSTR dst_file)
{
	FILE *file;

	file = _tfopen(dst_file, _T("wb"));

	if (!file)
		return FALSE;

	if (fwrite(src_buffer, 1, src_size, file) != src_size)
	{
		fclose(file);
		_tunlink(dst_file);
		return FALSE;
	}

	if (fclose(file) != 0)
	{
		_tunlink(dst_file);
		return FALSE;
	}

	return TRUE;
}

/****************************************************************************/

BOOL set_compat_flags(LPCTSTR exe_file)
{
	HKEY key;
	TCHAR flags[] = _T("$ DWM8And16BitMitigation 16BITCOLOR 8And16BitAggregateBlts DISABLEDWM HIGHDPIAWARE");

	if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, _T("SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\AppCompatFlags\\Layers"), 0, KEY_SET_VALUE | KEY_WOW64_64KEY, &key) != ERROR_SUCCESS)
		return FALSE;

	if (RegSetValueEx(key, exe_file, 0, REG_SZ, (LPCBYTE)flags, sizeof flags) != ERROR_SUCCESS)
	{
		RegCloseKey(key);
		return FALSE;
	}

	RegCloseKey(key);
	return TRUE;
}

/****************************************************************************/

BOOL delete_compat_flags(LPCTSTR exe_file)
{
	HKEY key;

	if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, _T("SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\AppCompatFlags\\Layers"), 0, KEY_SET_VALUE | KEY_WOW64_64KEY, &key) != ERROR_SUCCESS)
		return FALSE;

	if (RegDeleteValue(key, exe_file) != ERROR_SUCCESS)
	{
		RegCloseKey(key);
		return FALSE;
	}

	RegCloseKey(key);
	return TRUE;
}

/****************************************************************************/

BOOL get_gog_exe_path(LPTSTR exe_path, DWORD size)
{
	HKEY key;

	if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, _T("SOFTWARE\\GOG.com\\Games\\2086050016"), 0, KEY_QUERY_VALUE, &key) != ERROR_SUCCESS)
		return FALSE;

	size--;
	exe_path[size] = 0;
	size *= sizeof(TCHAR);

	if (RegQueryValueEx(key, "EXE", NULL, NULL, (LPBYTE)exe_path, &size) != ERROR_SUCCESS)
	{
		RegCloseKey(key);
		return FALSE;
	}

	RegCloseKey(key);
	return TRUE;
}

/* Main *********************************************************************/

int WINAPI _tWinMain(HINSTANCE instance, HINSTANCE previous_instance, LPTSTR command_line, int command_show)
{
	TCHAR exe_file[1024];
	TCHAR exe_path[1024];
	FILE *file;
	BOOL gog = FALSE;
	LPCTSTR name;
	LPCTSTR program_files;
	int index;
	TCHAR text[1024];
	TCHAR bak_file[1024];
	TCHAR new_file[1024];
	unsigned char *buffer;
	unsigned size;
	unsigned found;

	if (command_line[0])
	{
		if (_sntprintf(exe_file, 1000, _T("%s"), command_line) >= 1000)
			return error_message(_T("File name too long."));

		strip_quotes(exe_file);

		if (!_tfullpath(exe_path, exe_file, 1000))
			return error_message(_T("Failed to get full path name."));

		_tcscpy(exe_file, exe_path);
		strip_file_name(exe_path);

		if (_tchdir(exe_path) != 0)
			return error_message(_T("Specified path not found."));

		file = _tfopen(exe_file, _T("rb"));

		if (!file)
			return error_message(_T("Failed to open specified file."));

		if (get_gog_exe_path(exe_path, 1000))
			if (_tcsicmp(exe_file, exe_path) == 0)
				gog = TRUE;

		name = _tcsrchr(exe_file, _T('\\'));

		if (name)
			name++;
		else
			name = exe_file;
	}
	else
	{
		file = NULL;

		if (get_gog_exe_path(exe_path, 1000))
		{
			_tcscpy(exe_file, exe_path);
			strip_file_name(exe_path);

			if (_tchdir(exe_path) == 0)
			{
				file = _tfopen(exe_file, _T("rb"));

				if (file)
				{
					gog = TRUE;
					name = "SimCity 3000 Unlimited (GOG)";
				}
			}
		}

		if (!file)
		{
			program_files = _tgetenv(_T("ProgramFiles"));

			if (!program_files)
				program_files = _T("C:\\Program Files");

			if (_tchdir(program_files) != 0)
				return error_message(_T("Program files not found."));

			for (index = 0; editions[index].name; index++)
			{
				if (_sntprintf(exe_path, 1000, _T("%s\\Maxis\\%s\\%s\\%s"), program_files, editions[index].name, editions[index].exe_dir, editions[index].exe_file) >= 1000)
					continue;

				_tcscpy(exe_file, exe_path);
				strip_file_name(exe_path);

				if (_tchdir(exe_path) != 0)
					continue;

				file = _tfopen(exe_file, _T("rb"));

				if (!file)
					continue;

				name = editions[index].name;
				break;
			}

			if (!file)
				return error_message(_T("SimCity 3000 not found."));
		}
	}

	_sntprintf(text, 1000, _T("%s found.\n\n"), name);
	_sntprintf(bak_file, 1000, _T("%s.bak"), exe_file);
	_sntprintf(new_file, 1000, _T("%s.new"), exe_file);
	buffer = (unsigned char *)malloc(max_file_size + 1);

	if (!buffer)
		return error_message(_T("Failed to allocate memory."));

	size = fread(buffer, 1, max_file_size + 1, file);
	fclose(file);

	if (size == 0)
		return error_message(_T("Failed to read file."));

	if (size > max_file_size)
		return error_message(_T("File too large."));

	found = patch(buffer, size);

	if (!found)
	{
		if (!readable_file(bak_file))
		{
			_tcscat(text, _T("No resolutions found."));
			MessageBox(NULL, text, title, MB_OK);
			return 0;
		}

		_tcscat(text, _T("Restore from backup?"));

		if (MessageBox(NULL, text, title, MB_YESNO) != IDYES)
			return 0;

		if (!replace_file(bak_file, exe_file, NULL))
			return error_message(_T("Failed to restore file."));

		if (!gog)
			delete_compat_flags(exe_file);

		_trename(_T("Updater.bak"), _T("Updater"));
		MessageBox(NULL, _T("File successfully restored."), _T("Restore"), MB_ICONINFORMATION);
		return 0;
	}

	_tcscat(text, _T("Add desktop resolution?"));

	if (MessageBox(NULL, text, title, MB_YESNO) != IDYES)
		return 0;

	if (!write_file(buffer, size, new_file))
		return error_message(_T("Failed to write new file."));

	if (!replace_file(new_file, exe_file, bak_file))
	{
		_tunlink(new_file);
		return error_message(_T("Failed to replace file."));
	}

	set_compat_flags(exe_file);
	_trename(_T("Updater"), _T("Updater.bak"));
	MessageBox(NULL, _T("File successfully patched."), _T("Patch"), MB_ICONINFORMATION);
	return 0;
}
